/* This program creates a sitemap according to the open
   XML protocol proposed by Google. Optionally it can
   also generate an RSS 2.0 file (specify -rss in the
   command line).

   Input:
      Lines from stdin, containing the following tab-separated
      fields:
      - the publication date and time
      - the modification date and time
        (format of both: yyyy-mm-dd-hh:mm:ss)
      - an optional language code, according to ISO 639-1
        or ISO 639-2, optionally followed by an ISO 3166
        country code, optionally followed by an ISO 15924
        script code. The whole field is used for an optional
        selection via a command option (see below), and will
        be matched against what is found in <html lang=...>.
      - the path name (directories and file), relative to the
        current directory.

   Command line options:
      - --rss: without this option, the generated output is
           Google XML. With it, it is RSS 2.0.
        --lang=: the language code after this is used for
           selection. In the absence of --lang, everything
           is selected. With a language code, or several,
           only those will be selected from the lines.
           The separator may be a comma, colon or stop (,:.).
           Examples: --lang=en, --lang=ia,ina,ia-Grek
        An extra argument may override the default domain
          rudhar.com .
        --days=: the maximum number of days into the past
           the files of which should be included.

   Output:
      stdout, either XML or RSS.

   Late March 2021: I replaced my clumsy function to extract things
   from HTML with a more general solution, see getfhtml.h, which
   itself uses Google's Gumbo parser.

   26 January 2022: Error message when strptime detects an invalid
	date, so it is known where it occurs, and it can be corrected.
 */

#define _XOPEN_SOURCE 700 // See feature_test_macros(7)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#include "getfhtml.h"

struct attribs_s
{
   char *pubdate;
   char *moddate;
   /* Pubdate (if rss nonzero) or moddate (if rss is zero),
      recalculated to seconds since 1970-01-01, for easier
      comparison and calculation. */
   time_t pubmodseconds;

   char *langcod;
   char *pathnam;
   char *pgtitle;
};
struct parms_s
{
   /* 0 = generate Google XML sitemap; non-zero: generate RSS */
   short rss;
   /* Number of days into the past to select, older files will
      be skipped. Less than one means: select all. */
   short days_into_past;
   /* Time limit of the output, in second. Anything older than
      this point in time (assuming sorted input, newest on top)
      will be ignored. */
   time_t time_limit;
   /* What comes before the relative pathnames to make them absolute. */
   char *DomPref;
   /* Pointer to NULL terminated vector of char pointers.
      Langsel is malloc'ed and must be freed. The pointers
      of the vector point into an argv member. */
   char **LangSel;
};

static int AnalyseParms (int argc, char **argv, struct parms_s *p_parms);
static int AnalyseLine (char *linebuf, struct attribs_s *p_attribs, int rss);
static int SelectLang  (struct parms_s *p_parms, struct attribs_s *p_attribs);
static int WriteFileSitemap (
   FILE *fpo, struct parms_s *p_parms, struct attribs_s *p_attribs);
static int WriteFileRssFeed (
   FILE *fpo, struct parms_s *p_parms, struct attribs_s *p_attribs);
static int Praefix (FILE *fpo, int rss, char *DomPref);
static int Postfix (FILE *fpo, int rss);

static char *DomainPrefix = "https://rudhar.com";
// static char *DomainPrefix = "https://rhar.info";
static char ChFreq[128];
static char timestamp[128];

#define LAGFAC  (double)0.75
#define LAGHOUR (long)(LAGFAC*60.0*60.0)

int main (int argc, char **argv)
{
   static char linebuf[1024];
   struct attribs_s attribs;
   struct parms_s   parms;
   FILE *fpi = stdin;
   FILE *fpo = stdout;

   AnalyseParms(argc, argv, &parms);
   Praefix(fpo, parms.rss, parms.DomPref);

   while (fgets(linebuf, sizeof linebuf, fpi))
   {
      struct stat st;

      AnalyseLine(linebuf, &attribs, parms.rss);

      /* Accept only regular files, skip directories. */
      if (stat(attribs.pathnam, &st) != 0 || !(st.st_mode & S_IFREG))
         continue;

      if (!SelectLang(&parms, &attribs))
         continue;

      /* Assuming sorted dates, newest on top, in case of a period
         limited via the command line, ignore the rest of the lines. */
      if (parms.time_limit &&
            difftime(attribs.pubmodseconds, parms.time_limit) < 0)
         break;

      if (parms.rss)
         WriteFileRssFeed(fpo, &parms, &attribs);
      else
         WriteFileSitemap(fpo, &parms, &attribs);
   }
   Postfix(fpo, parms.rss);
   free(parms.LangSel);

   return 0;
}

static int AnalyseParms (int argc, char **argv, struct parms_s *p_parms)
{
   char *langopt = "--lang=";
   char *daysopt = "--days=";
   p_parms->rss = 0;
   p_parms->days_into_past = 0;
   p_parms->LangSel = NULL;
   p_parms->DomPref = DomainPrefix;

   while (argc > 1 && argv[1])
   {
      if (strcmp(argv[1], "--rss") == 0)
         p_parms->rss = 1;
      else if (strncmp(argv[1], langopt, strlen(langopt)) == 0)
      {
         char *p, *q;
         int breakcond;
         size_t vectlen = 0;

         p_parms->LangSel = realloc(p_parms->LangSel,
                                    ++vectlen * sizeof p_parms->LangSel);
         p_parms->LangSel[vectlen - 1] = NULL;
         p = argv[1] + strlen(langopt);
         while (q = p + strcspn(p, ",:."), q - p)
         {
            p_parms->LangSel = realloc(p_parms->LangSel,
                                       ++vectlen * sizeof p_parms->LangSel);
            breakcond = !*q;
            *q = '\0';
            p_parms->LangSel[vectlen - 2] = p;
            if (breakcond)
               break;
            p = q + 1;
         }
         p_parms->LangSel[vectlen - 1] = NULL;
      }
      else if (strncmp(argv[1], daysopt, strlen(daysopt)) == 0)
      {
         p_parms->days_into_past = atoi(argv[1] + strlen(daysopt));
      }
      else if (strncmp(argv[1], "--", 2) != 0)
      {
         char *p;
         p_parms->DomPref = argv[1];
         /* No closing slash, will be added later if necessary. */
         p = p_parms->DomPref + strlen(p_parms->DomPref);
         if (p > p_parms->DomPref && p[-1] == '/')
            *p = '\0';
      }
      else
         ; /* Unknown options are silently ignored */

      argc--, argv++;
   }

   p_parms->time_limit = 0;
   if (p_parms->days_into_past)
      p_parms->time_limit = time(NULL) - p_parms->days_into_past * 3600L * 24;

   return 0;
}

static int AnalyseLine (char *linebuf, struct attribs_s *p_attribs, int rss)
{
   char *p, *q;

   p_attribs->pubdate = NULL;
   p_attribs->moddate = NULL;
   p_attribs->langcod = NULL;
   p_attribs->pathnam = NULL;

   p = linebuf;
   if ((q = strchr(p, ';')))
   {
      *q = '\0';
      p_attribs->pubdate = p;
      p = q + 1;
   }
   if ((q = strchr(p, ';')))
   {
      *q = '\0';
      p_attribs->moddate = p;
      p = q + 1;
   }
   if ((q = strchr(p, ';')))
   {
      *q = '\0';
      p_attribs->langcod = p;
      p = q + 1;
   }
   if ((q = strchr(p, '\n')))
   {
      *q = '\0';
      p_attribs->pathnam = p;
      p = q + 1;
   }
   {
      struct tm tm;
      char *retval;
      char *date_to_consider = rss ? p_attribs->pubdate : p_attribs->moddate;

      memset(&tm, '\0', sizeof tm);
      tm.tm_isdst = -1; /* Unknown if daylight saving time in effect */
      retval = strptime(date_to_consider, "%Y-%m-%dT%H:%M:%S", &tm);
      if (!retval)
         fprintf(stderr, "Date-time conversion error on %s for %s\n",
            date_to_consider, p_attribs->pathnam);
      p_attribs->pubmodseconds = mktime(&tm);
   }

   return 0;
}

static int SelectLang (struct parms_s *p_parms, struct attribs_s *p_attribs)
{
   /* If no language code of codes were specified in the command
      line, or they were empty, that means "select all", so return
      "yes". */
   if (!p_parms->LangSel ||
       !p_parms->LangSel[0] ||
       !p_parms->LangSel[0][0])
      return 1;

   for (char **l = p_parms->LangSel; *l; l++)
   {
      if (strcmp(*l, p_attribs->langcod) == 0)
         return 1;
   }
   /* There were one or more codes in the command line, but
      none of them matches. Then don't select the line & path. */
   return 0;
}

static int WriteFileSitemap (
   FILE *fpo, struct parms_s *p_parms, struct attribs_s *p_attribs)
{
   char *path;
   struct tm *tm;
   time_t now;
   time_t tdif;

   fprintf(fpo, "\t<url>\n");

   tm = gmtime(&(p_attribs->pubmodseconds));
   strftime(timestamp, sizeof timestamp, "%Y-%m-%dT%H:%M:%SZ", tm);
   now = time(NULL);
   tdif = now - p_attribs->pubmodseconds;

   // printf("%ld - %ld = %ld\n", now, st.st_mtime, tdif);

   if ((long)tdif < LAGHOUR)
      strcpy(ChFreq, "hourly");
   else if ((long)tdif < LAGHOUR * 24)
      strcpy(ChFreq, "daily");
   else if ((long)tdif < LAGHOUR * 24 * 7)
      strcpy(ChFreq, "weekly");
   else if ((long)tdif < LAGHOUR * 24 * 30)
      strcpy(ChFreq, "monthly");
   else if ((long)tdif < LAGHOUR * 24 * 365)
      strcpy(ChFreq, "yearly");
   else
      strcpy(ChFreq, "never");

   path = p_attribs->pathnam;
   if (strncmp(path, "./", 2) == 0)
      path += 2;

   /* Example:
      <loc>https://www.example.com/</loc>
      <lastmod>2005-01-01</lastmod>
      <changefreq>monthly</changefreq>
      <priority>0.8</priority> */
   fprintf(fpo, "\t\t<loc>%s/%s</loc>\n", p_parms->DomPref, path);
   fprintf(fpo, "\t\t<lastmod>%s</lastmod>\n", timestamp);
   fprintf(fpo, "\t\t<changefreq>%s</changefreq>\n", ChFreq);

   fprintf(fpo, "\t</url>\n");

   return 0;
}

static int WriteFileRssFeed (
   FILE *fpo, struct parms_s *p_parms, struct attribs_s *p_attribs)
{
   char *path;
   struct tm *tm;
   char *p;
   char *FoundTitle = NULL;

   tm = gmtime(&(p_attribs->pubmodseconds));
   strftime(timestamp, sizeof timestamp, "%a, %d %b %Y %H:%M:%S GMT", tm);

   fprintf(fpo, "\t\t<item>\n");

   path = p_attribs->pathnam;
   if (strncmp(path, "./", 2) == 0)
      path += 2;

   FoundTitle = gfh_GetFromPath(p_attribs->pathnam, 1 << 9, "title", NULL);
   if (FoundTitle && strlen(FoundTitle))
      fprintf(fpo, "\t\t\t<title><![CDATA[%s]]></title>\n", FoundTitle);
   else
      fprintf(fpo, "\t\t\t<title><![CDATA[%s/%s]]></title>\n", p_parms->DomPref, path);
   free(FoundTitle);

   fprintf(fpo, "\t\t\t<link>%s/%s</link>\n", p_parms->DomPref, path);
   fprintf(fpo, "\t\t\t<pubdate>%s</pubdate>\n", timestamp);
   fprintf(fpo, "\t\t</item>\n");

   return 0;
}

static int Praefix (FILE *fpo, int rss, char *DomPref)
{
   if (rss)
   {
      struct tm *t;
      time_t now;

      now = time(NULL);
      t = gmtime(&now);
      strftime(timestamp, sizeof timestamp, "%a, %d %b %Y %H:%M:%S GMT", t);

      fprintf(fpo, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
      fprintf(fpo, "<rss version=\"2.0\">\n");
      fprintf(fpo, "\t<channel>\n");
      fprintf(fpo, "\t\t<title>rudhar.com</title>\n");
      fprintf(fpo, "\t\t<description>Words, music, opinions and weird colours</description>\n");
      fprintf(fpo, "\t\t<link>%s</link>\n", DomPref);
      fprintf(fpo, "\t\t<lastBuiltDate>%s</lastBuiltDate>\n", timestamp);
   }
   else
   {
      fprintf(fpo, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
      fprintf(fpo, "<urlset xmlns=\"https://www.sitemaps.org/schemas/sitemap/0.9\">\n");
   }
   return 0;
}

static int Postfix (FILE *fpo, int rss)
{
   if (rss)
   {
      fprintf(fpo, "\t</channel>\n");
      fprintf(fpo, "</rss>\n");
   }
   else
   {
      fprintf(fpo, "</urlset>\n");
   }

   return 0;
}
